/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/math/fpu.c: Functions for configuring the floating-point environment.
 */

#include "src/base.h"
#include "src/math/fpu.h"

/*************************************************************************/
/*************************************************************************/

void fpu_configure(void)
{
    #if defined(__GNUC__) || defined(__clang__)

        #if defined(SIL_ARCH_ARM)

            #ifndef SIL_PLATFORM_ANDROID
                uint32_t fpscr;
                __asm__("vmrs %0, fpscr" : "=r" (fpscr));
                fpscr &= ~(3<<22);  // Round to nearest value.
                fpscr |= 1<<24;     // Flush denormals to zero.
                #ifndef DEBUG
                    fpscr &= ~(0x9F<<8);  // Disable exceptions.
                #endif
                __asm__("vmsr fpscr, %0" : /* no outputs */ : "r" (fpscr));
            #else
                /* Apparently the Android NDK compiler doesn't understand
                 * VMRS/VMSR (which go all the way back to VFPv2, so no
                 * NEON support isn't an excuse!), so we have to encode
                 * them manually. */
                __asm__ volatile(
                    ".word 0xEEF10A10\n"  // vmrs r0, fpscr
                    "orr r0, r0, #(1<<24)\n"
                    "and r0, r0, #~(3<<22)\n"
                    #ifndef DEBUG
                        "and r0, r0, #~(0x9F<<8)\n"
                    #endif
                    ".word 0xEEE10A10\n"  // vmsr fpscr, r0
                    : /* no outputs */
                    : /* no inputs */
                    : "r0");
            #endif  // SIL_PLATFORM_ANDROID

        #elif defined(SIL_ARCH_MIPS)

            uint32_t fpucr;
            __asm__("cfc1 %[fpucr], $31" : [fpucr] "=r" (fpucr));
            fpucr &= ~3;     // Round to nearest value.
            fpucr |= 1<<24;  // Flush denormals to zero.
            #ifndef DEBUG
                fpucr &= ~(0x1F << 7);  // Disable all exceptions.
            #endif
            __asm__("ctc1 %[fpucr], $31" : : [fpucr] "r" (fpucr));

        #elif defined(SIL_ARCH_X86)

            uint32_t mxcsr;
            __asm__("stmxcsr %0" : "=m" (mxcsr));
            mxcsr &= ~(3<<13);  // RC (00 = round to nearest)
            mxcsr |= 1<<6;      // DAZ
            #ifndef DEBUG
                mxcsr |= 1<<7       // EM_INVALID
                      |  1<<8       // EM_DENORMAL
                      |  1<<9       // EM_ZERODIVIDE
                      |  1<<10      // EM_OVERFLOW
                      |  1<<11      // EM_UNDERFLOW
                      |  1<<12;     // EM_PRECISION
            #endif
            __asm__("ldmxcsr %0" : /* no outputs */ : "m" (mxcsr));

        #else  // !ARM && !MIPS && !X86

            #warning Cannot set floating-point mode flags on this CPU!

        #endif

    #else  // !__GNUC__ && !__clang__

        #warning Cannot set floating-point mode flags with this compiler!

    #endif
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
